# 3rd party imports
# Local imports
from .utils.stellar import *


def get_account_balances(account_address):
    """
    This method is used to fetch all the balances from the given account address.
    :param str account_address: Account address to be evaluated.
    :return: Returns a list containing the account balances structured in the following manner
    : [['token1', amount], ['token2', amount]]
    :rtype: list of (str, str)
    """
    address = get_address_details_from_network(account_address)
    if address is None:
        return None

    balances = []
    for balance in address.balances:
        token_code = balance.get('asset_code', 'XLM')
        token_balance = float(balance.get('balance'))
        balances.append([token_code, token_balance])
    return balances


def get_account_payments(account_address):
    """
    This method is used to fetch all the payments from the given account address.
    :param str account_address: Account address to be evaluated.
    :return: Returns a JSON string with the payments. TODO: The string must be parsed
    :rtype: str
    """
    address = get_address_details_from_network(account_address)
    if address is None:
        return None

    return address.payments()


def get_account_transactions(account_address):
    """
    This method is used to fetch all the transactions from the given account address.
    :param str account_address: Account address to be evaluated.
    :return: Returns a JSON string with the transactions. TODO: The string must be parsed
    :rtype: str
    """
    address = get_address_details_from_network(account_address)
    if address is None:
        return None

    return address.transactions()



# System imports
import json
from json import JSONDecodeError


def is_float_str(string):
    """
    Checks if a given str can be successfully converted to a float value.
    :param str string: String to be evaluated.
    :return: Returns true if the string is float convertible and false otherwise.
    :rtype: bool
    """
    try:
        float(string)
        return True
    except ValueError:
        return False


def is_int_str(string):
    """
    Checks if a given str can be successfully converted to an integer value.
    :param str string: String to be evaluated.
    :return: Returns true if the string is integer convertible and false otherwise.
    :rtype: bool
    """
    try:
        int(string)
        return True
    except ValueError:
        return False


def is_in_range(val, range_min, range_max):
    """
    Checks if the specified value is within the given range.
    :param (int, float) val: Value to be evaluated.
    :param (int, float) range_min: Range lower value.
    :param (int, float) range_max: Range higher value.
    :return: Returns true if the value is within the range and false otherwise.
    :rtype: bool
    """
    return True if range_min <= val <= range_max else False


def is_successful_http_status_code(status_code):
    """
    Checks if a given HTTP status code is successfull. For this to happen
    this value should be between 200 and 299 (2XX)
    :param int status_code: Status code to be evaluated.
    :return: Returns true if the status code is successful and false otherwise.
    :rtype: bool
    """
    return is_in_range(status_code, 200, 299)


def decode_json_content(content):
    """
    Decodes a given string content to a JSON object
    :param str content: content to be decoded to JSON.
    :return: A JSON object if the string could be successfully decoded and None otherwise
    :rtype: json or None
    """
    try:
        return json.loads(content) if content is not None else None
    except JSONDecodeError:
        print("The given content could not be decoded as a JSON file")
        return None



# 3rd party imports
from stellar_base.keypair import Keypair
# Local imports
from .constants import *
from .utils.user_input import *
from .utils.file import *
from .utils.generic import *
from .utils.stellar import *


class CliSession:
    """
    
    Attributes
    ----------
    account_name : str
        Name of the Stellar account used on the CLI session.
    account_address : str
        Address of the Stellar account used on the CLI session.
    account_seed : str
        Secret seed of the Stellar account used on the CLI session. The seed
        is a string that can be used to decode both private and public key of the account.
    """

    def __init__(self, account_name, account_address, account_seed):

        self.account_name: str = account_name
        self.account_address: str = account_address
        self.account_seed: str = account_seed

    def store_account_in_config_file(self):
        if not _config_file_exists():
            password = password_input("Please insert an encryption/decryption password "
                                      "to protect the configuration file")
            if password != password_input("Please insert the encryption/decryption password again"):
                print("The inserted passwords are distinct. The configuration file could not be created")
                return
            configs_json = CONFIG_FILE_EMPTY_JSON
        else:
            password = password_input("Please insert the decryption password of the configuration file")
            configs_json = _config_file_load(password)
            if configs_json is None:
                print("The configuration file is either empty, could not be decrypted or is an invalid JSON file.")
                return

        configs_json[JSON_ACCOUNTS_TAG].append({
            JSON_ACCOUNT_NAME_TAG: self.account_name,
            JSON_ACCOUNT_ADDRESS_TAG: self.account_address,
            JSON_ACCOUNT_SEED_TAG: self.account_seed})

        _config_file_write(json.dumps(configs_json), password)

    def fetch_valid_seed(self):
        if self.account_seed is None \
                or not is_seed_valid(self.account_seed) \
                or not is_seed_matching_address(self.account_seed, self.account_address):
            self.account_seed = self.ask_for_user_seed("Either no seed was found for this CLI session account, "
                                                       "the seed for this CLI session account is invalid or "
                                                       "the seed does match the current CLI session account address. "
                                                       "No transaction can be made without a valid seed. Please "
                                                       "insert your seed to process the transaction")
        return self.account_seed

    def ask_for_user_seed(self, msg):
        seed = password_input(msg)
        if not is_seed_valid(seed):
            print('The given seed is invalid')
            return None

        if not is_seed_matching_address(seed, self.account_address):
            print('The given seed does not match with the address of the current CLI session')
            return None

        if yes_or_no_input('Do you want to save the seed for this CLI session account?') == USER_INPUT_YES:
            self.account_seed = seed

        return seed

    def to_str(self):
        return 'Account Name: {}, Account Address: {}'.format(self.account_name, self.account_address)


def init_cli_session():
    """
    This function initializes a CLI session. There are three alternatives to initialize a session:
    1: Initialize from an account stored on the configuration file.
    2: Initialize from an existent account which is not stored on the configuration file.
    3: Initialize from a newly created Stellar account.
    :return: Returns the initialized CLI session or None if something went wrong.
    :rtype: CliSession or None
    """
    if not _config_file_exists():
        print("No configuration file was found.")
    elif yes_or_no_input('Do you want to use an account from the configuration file?') == USER_INPUT_YES:
        return _init_session_from_conf_file()

    if yes_or_no_input('Do you want to use an existent account?') == USER_INPUT_YES:
        return _init_session_from_existent_account()

    if yes_or_no_input('Do you want to create a new account?') == USER_INPUT_YES:
        return _init_session_from_new_account()

    return None


def _init_session_from_conf_file():
    """
    Initializes a session from an account stored on the configuration file. This function assumes that the
    configuration file already exists. This method can fail in the following situations:
    1: The configuration file is empty.
    2: The configuration file could not be decrypted (for example if a wrong decryption password was given).
    3: The configuration file is not a valid JSON file.
    4: The configuration file does not have the 'accounts' JSON key, or there are no accounts stored on it.
    5: The index of the configuration file account specified by the user is invalid.
    :return: Returns the initialized CLI session or None if something went wrong.
    :rtype: CliSession or None
    """
    configs_json, conf_file_password = _config_file_load()
    if configs_json is None:
        print("The configuration file is either empty, could not be decrypted or is an invalid JSON file.")
        return None

    if not _config_file_is_valid_content(configs_json):
        print("The content of the configuration file is invalid")
        return None

    n_accounts_found = _config_file_get_number_of_accounts(configs_json)
    if n_accounts_found <= 0:
        print("No accounts were found on the configuration file.")
        return None

    _config_file_print_accounts(configs_json)
    account_n = int_input('Which account do you want to use? (specify the index)') - 1
    if account_n < 0 or account_n >= n_accounts_found:
        print("Specified account index is invalid")
        return None

    name = configs_json[JSON_ACCOUNTS_TAG][account_n][JSON_ACCOUNT_NAME_TAG]
    address = configs_json[JSON_ACCOUNTS_TAG][account_n][JSON_ACCOUNT_ADDRESS_TAG]
    seed = configs_json[JSON_ACCOUNTS_TAG][account_n].get(JSON_ACCOUNT_SEED_TAG, None)
    return CliSession(name, address, seed)


def _init_session_from_existent_account():
    """
    Initializes a session from an account which is not stored on the configuration file. This method fails
    if the given Stellar address is invalid.
    :return: Returns the initialized CLI session or None if something went wrong.
    :rtype: CliSession or None
    """
    address = safe_input("Please insert the address of the existent account")
    if not is_address_valid(address):
        print("The specified address is not a valid Stellar address.")
        return None

    cli_session = CliSession(None, address, None)
    if _config_file_is_to_update():
        if yes_or_no_input("Do you want to give a name to the new account to be created?") == USER_INPUT_YES:
            cli_session.account_name = safe_input("Please insert the account name")
        cli_session.store_account_in_config_file()
    return cli_session


def _init_session_from_new_account():
    """
    Initializes a session from a newly created account. A Stellar keypair is randomly generated.
    :return: Returns the initialized CLI session.
    :rtype: CliSession
    """
    account_name = None
    if yes_or_no_input("Do you want to give a name to the new account to be created?") == USER_INPUT_YES:
        account_name = safe_input("Please insert the account name")

    keypair = Keypair.random()
    address = keypair.address().decode()
    seed = keypair.seed().decode()

    cli_session = CliSession(account_name, address, seed)
    if _config_file_is_to_update():
        if yes_or_no_input("Do you want to save the seed of the new account in the encrypted configuration file") == \
                USER_INPUT_YES:
            cli_session.store_account_in_config_file()
        else:
            seed_file = safe_input("Specify a file to which the seed of the new created account should be saved. "
                                   "Notice that the seed will be saved in plain text on the specified file. "
                                   "Please store it in a safe place. If anyone can reach the seed it can fully "
                                   "control your account")
            write_file(seed_file, seed)
            cli_session.store_account_in_config_file()
    return cli_session


def _config_file_print_accounts(configs_json):
    """
    Prints the accounts found on the configuration file.
    :param json configs_json: Configuration file JSON decoded data.
    """
    accounts = _config_file_get_accounts(configs_json)
    print('The following {} Stellar accounts were found on the configuration file:'.format(len(accounts)))
    for i, account in enumerate(accounts):
        print('[{}] Account Name: {}, Account Address: {}'.format(
            i+1, account[JSON_ACCOUNT_NAME_TAG], account[JSON_ACCOUNT_ADDRESS_TAG]))
    print('')


def _config_file_exists():
    """
    Checks if the configuration file exists.
    :return: Returns True if the configuration file exists and False otherwise.
    :rtype: bool
    """
    if os.path.isfile(DEFAULT_CONFIG_FILE):
        return True
    return False


def _config_file_load(password=None):
    """
    Requests a password to decrypt the configuration file, decrypts it and returns the decoded JSON content
    together with the decryption password. If the file cannot be opened or decrypted or if its content cannot
    be decoded as JSON this method returns None.
    :param str password: Password to decrypt the file. If no password is given one is requested from the user.
    :return: In case of success it returns.
    :rtype: (json, str) or (None, None)
    """
    if password is None:
        password = password_input("Please insert the configuration file decryption password")
    configs_content = decode_json_content(load_encrypted_file(DEFAULT_CONFIG_FILE, password))
    return (None, None) if configs_content is None else (configs_content, password)


def _config_file_write(config_json, password):
    """
    Encrypts the specified content with the given password and writes it to the configuration file.
    :param json config_json: Content to be stored on the configuration file.
    :param password: Encryption password,
    """
    file = DEFAULT_CONFIG_FILE
    succeeded = write_encrypted_file(file, config_json, password)
    if succeeded:
        print("Configuration file successfully stored in: {}".format(file))


def _config_file_is_to_update():
    """
    Ask the user if the configuration file should be updated or not.
    :return: Returns True if the user wants to update the configuration file and False otherwise.
    :rtype: bool
    """
    if yes_or_no_input("Do you want to save the account on the configuration file?") == USER_INPUT_YES:
        return True
    return False


def _config_file_is_valid_content(configs_json):
    """
    Checks if the content of the configuration file is valid. For this to be true an 'accounts' JSON key
    must be found.
    :param json configs_json: Configuration file JSON decoded data.
    :return: Returns True if the content is valid and False otherwise.
    :rtype: bool
    """
    if JSON_ACCOUNTS_TAG in configs_json:
        return True
    return False


def _config_file_get_number_of_accounts(configs_json):
    """
    Returns the number of accounts stored on the configuration file.
    :param json configs_json: Configuration file JSON decoded data.
    :return: Returns the number of accounts stored on the configuration file.
    :rtype: int
    """
    return len(configs_json[JSON_ACCOUNTS_TAG])


def _config_file_get_accounts(configs_json):
    """
    Returns the accounts stored on the configuration file. It assumes that the given configuration JSON
    content is valid (which means that it must have an 'accounts' JSON key).
    :param json configs_json: Configuration file JSON decoded data.
    :return: Returns the accounts stored on the configuration file.
    :rtype: json
    """
    return configs_json[JSON_ACCOUNTS_TAG]



import base64
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.fernet import InvalidToken


def encrypt(content, password):
    """
    Method to encrypt the given content based on the specified password.
    :param bytearray content: Content to be encrypted.
    :param str password: Encryption password.
    :return: Returns a byte array with the encrypted content.
    :rtype: bytearray
    """
    f = Fernet(password2cryptographic_key(password))
    return f.encrypt(content)


def decrypt(content, password):
    """
    Method to decrypt the given content based on the specified password.
    :param bytearray content: Content to be decrypted.
    :param str password: Decryption password.
    :return: Returns a byte array with the decrypted content.
    :rtype: bytearray
    """

    f = Fernet(password2cryptographic_key(password))
    try:
        decrypted_content = f.decrypt(content)
    except InvalidToken:
        print("The file could not be decrypted. Please check if the password is correct")
        return None
    return decrypted_content


def password2cryptographic_key(password):
    """
    This method generates a key, based on the specified password, to be used during the
    encryption and decryption procedures.
    :param str password: Password to be used during the key generation.
    :return: Returns the generated key.
    """
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=bytes("0"*16, 'utf-8'),  # os.urandom(16) TODO: Check salt influence
        iterations=100000,
        backend=default_backend())

    key = kdf.derive(password.encode())  # derive key from the user password
    return base64.urlsafe_b64encode(key)



# System imports
import getpass
# Local imports
from .generic import *

USER_INPUT_YES = 'y'
USER_INPUT_NO = 'n'


def int_input(msg):
    """
    This methods should be used to request an integer value from the user. This
    method will not return until the user inputs a valid integer value.
    :param str msg: Message to be displayed to the user.
    :return: Returns the integer value inserted by the user.
    :rtype: int
    """
    value = safe_input(msg)
    while not is_int_str(value):
        print('Input must be a valid integer value')
        value = safe_input(msg)
    return int(value, base=10)


def float_input(msg):
    """
    This methods should be used to request a float value from the user. This
    method will not return until the user inputs a valid float value.
    :param str msg: Message to be displayed to the user.
    :return: Returns the float value inserted by the user.
    :rtype: float
    """
    value = safe_input(msg).replace(',', '.')
    while not is_float_str(value):
        print('Input must be a valid float value')
        value = safe_input(msg).replace(',', '.')
    return float(value)


def yes_or_no_input(msg):
    """
    This methods should be used to request an Yes/No input from the user. This
    method will not return until the user inputs a valid answer (USER_INPUT_YES or
    USER_INPUT_NO).
    :param str msg: Message to be displayed to the user.
    :return: Returns USER_INPUT_YES if the answer is affirmative and USER_INPUT_NO otherwise.
    :rtype: str
    """
    full_msg = '{} ({}/{})'.format(msg, USER_INPUT_YES, USER_INPUT_NO)
    answer = safe_input(full_msg).lower()
    while answer not in [USER_INPUT_YES, USER_INPUT_NO]:
        answer = safe_input(full_msg).lower()
    return answer


def password_input(msg):
    """
    This methods should be used to request a password input from the user. With
    this method the password inserted by the user will not be displayed on the
    screen. It also exits gracefully in case of a KeyboardInterrupt signal (ctr+c)
    being received.
    :param str msg: Message to be displayed to the user.
    :return: Returns the password inserted by the user.
    :rtype: str
    """
    try:
        return getpass.getpass('{}: '.format(msg))
    except KeyboardInterrupt:
        print('\n\nExiting..')
        raise SystemExit


def safe_input(msg):
    """
    This method extends the system built input() method in order to exit gracefully
    in case of a KeyboardInterrupt signal (ctr+c) being received.
    :param str msg: Message to be displayed to the user.
    :return: Returns the input inserted by the user.
    :rtype: str
    """
    try:
        return input('{}: '.format(msg))
    except KeyboardInterrupt:
        print('\n\nExiting..')
        raise SystemExit



# System imports
import os
# Local imports
from .cryptography import *

FILE_MODE_READ = 'r'
FILE_MODE_READ_BINARY = 'rb'
FILE_MODE_WRITE = 'w'
FILE_MODE_WRITE_BINARY = 'wb'


def write_file(filename, content, write_as_binary=False):
    """
    Writes the given content to the specified file.
    :param str filename: File to which the content should be written.
    :param str content: Content to be written.
    :param bool write_as_binary: Flag that specifies if the content must be written in binary form or not.
    This parameter is optional and by default files are not written is binary form.
    :return: Returns True in case of success and false otherwise.
    :rtype: bool
    """
    opening_mode = FILE_MODE_WRITE if not write_as_binary else FILE_MODE_WRITE_BINARY
    try:
        file = open(filename, opening_mode)
        file.write(content)
        file.close()
        return True
    except (OSError, IOError):
        print("There was a problem opening/writing the file: {}".format(filename))
        return False


def load_file(filename, read_as_binary=False):
    """
    Loads the specified file to memory.
    :param str filename: File to be loaded.
    :param bool read_as_binary: Flag that specifies if the file must be loaded in binary form or not.
    This parameter is optional and by default files are not loaded is binary form.
    :return: Returns the content of the file or None if the file could not be read.
    :rtype: str or bytearray or None
    """
    if os.path.isfile(filename):
        opening_mode = FILE_MODE_READ if not read_as_binary else FILE_MODE_READ_BINARY
        try:
            with open(filename, opening_mode) as file_content:
                return file_content.read()
        except (OSError, IOError):
            print("There was a problem opening/reading the file: {}".format(filename))
            return None
    return None


def write_encrypted_file(filename, content, password):
    """
    Writes the given content to the specified file and encrypts it with the specified password.
    :param str filename: File to which the content should be written.
    :param str content: Content to be written.
    :param str password: Password to encrypt the file.
    :return: Returns True in case of success and false otherwise.
    :rtype: bool
    """
    if not _is_valid_password(password):
        print('A valid password must be given to decrypt the file')
        return False

    encrypted_content = encrypt(content.encode(), password)
    return write_file(filename, encrypted_content, write_as_binary=True)  # Encrypted files are stored as binary


def load_encrypted_file(filename, password, read_as_binary=False):
    """
    Loads the specified file to memory and decrypts it with the specified password.
    :param str filename: File to be loaded.
    :param str password: Password to decrypt the file.
    :param bool read_as_binary: Flag that specifies if the file must be loaded in binary form or not.
    This parameter is optional and by default files are not loaded is binary form.
    :return: Returns the content of the file or None if the file could not be read.
    :rtype: str or bytearray or None
    """
    if not _is_valid_password(password):
        print('A valid password must be given to decrypt the file')
        return None

    file_content = load_file(filename, read_as_binary=True)  # Encrypted files are stored as binary
    if file_content is None:
        return None

    decrypted_content = decrypt(file_content, password)
    if decrypted_content is None:
        return None

    return decrypted_content if read_as_binary else decrypted_content.decode()


def _is_valid_password(password):
    """
    Evaluates if a given password is valid. Its length must be higher than 0.
    :param str password: Password to be evaluated.
    :return: Returns True if the password is valid and False otherwise.
    :rtype: bool
    """
    if password is None or len(password) == 0:
        return False
    return True



# System imports
import requests
import base64
# 3rd party imports
from stellar_base.operation import *
from stellar_base.builder import Builder
from stellar_base.stellarxdr import Xdr
from stellar_base.stellarxdr import StellarXDR_const
from xdrlib import Error as XDRError
# Local imports
from .utils.stellar import *
from .utils.user_input import *
from .constants import *
from .utils.generic import *
from .stellar_operations import *


def create_new_account(source_account_address, source_account_seed, new_account_address, amount, transaction_memo=''):
    """
    This method creates a new Stellar account. For this to be done a certain amount
    of XLM must be transferred from a source account to the new account.
    :param str source_account_address: Address of the account from which the funds to
    create the new account will be retrieved.
    :param str source_account_seed: Seed of the account from which the funds to
    create the new account will be retrieved.
    :param str new_account_address: Address of the new account.
    :param str amount: XLM amount to transfer to the new account.
    :param str transaction_memo: Text memo to be included in Stellar transaction. Maximum size of 28 bytes.
    """
    if not is_seed_matching_address(source_account_seed, source_account_address):
        print("The new account could not be created. Either the given source account address or source"
              "account seed are invalid or they do not match.")
        return

    if not is_address_valid(new_account_address):
        print('The given account address is invalid')
        return

    if yes_or_no_input('To create the new account, a payment of {} XLM will be done '
                       'to the following address {}. Are you sure you want to proceed?'
                       .format(amount, new_account_address)) == USER_INPUT_NO:
        return

    op = create_account_creation_op(new_account_address, amount, source_account_address)
    response = submit_operations(source_account_seed, op, transaction_memo)
    # process_server_payment_response(response) # TODO: Parse response


def fund_using_friendbot(account_address):
    """
    This method is used to request the Stellar Friendbot to fund the given account
    address. This will only work on the Stellar testnet.
    :param str account_address: Account address to be funded.
    :return: Returns a string with the result of the fund request.
    :rtype: str
    """
    if not is_address_valid(account_address):
        return 'The given account address is invalid.'

    try:
        r = requests.get('{}/friendbot?addr={}'.format(STELLAR_HORIZON_TESTNET_URL, account_address))
        return 'Successful transaction request' if is_successful_http_status_code(r.status_code) \
            else 'Failed transaction request (Maybe this account was already funded by Friendbot). Status code {}'.\
            format(r.status_code)
    except requests.exceptions.ConnectionError:
        return "A connection error occurred (Please check your Internet connection)"


def send_payment(source_account_address, source_account_seed, destination_account_address,
                 token_code, amount, token_issuer=None, transaction_memo=''):
    """
    This method is used to send a transaction of the specified token to a given address.
    :param str source_account_address: Address of the account from which the funds will be sent.
    :param str source_account_seed: Seed of the account from which the funds will be sent.
    :param str destination_account_address: Destination address.
    :param str amount: Amount to be sent.
    :param str token_code: Code of the token to be sent.
    :param str token_issuer: Issuer of the token to be sent. It can be None when dealing with native asset (XLM).
    :param str transaction_memo: Text memo to be included in Stellar transaction. Maximum size of 28 bytes.
    """
    if not is_seed_matching_address(source_account_seed, source_account_address):
        print("The payment could not be finalized. Either the given source account address or source"
              "account seed are invalid or they do not match.")
        return

    if not is_address_valid(destination_account_address):
        print('The given destination address is invalid')
        return
    if destination_account_address == source_account_address:
        print('Sending payment to own address. This is not allowed')
        return
    if token_issuer is not None and not is_address_valid(token_issuer):
        print('The given token issuer address is invalid')
        return
    if not is_transaction_text_memo_valid(transaction_memo):
        print('The maximum size of the text memo is {} bytes'.format(STELLAR_MEMO_TEXT_MAX_BYTES))
        return

    if yes_or_no_input('A payment of {} {} will be done to the following address {}. Are you sure you want to proceed?'
                       .format(amount, token_code, destination_account_address)) == USER_INPUT_NO:
        return

    operations = [create_payment_op(destination=destination_account_address,
                                    amount=amount,
                                    asset_code=token_code,
                                    asset_issuer=token_issuer,
                                    source=source_account_address)]
    response = submit_operations(source_account_seed, operations, transaction_memo)
    process_server_payment_response(response)


def send_path_payment(source_account_address, source_account_seed, destination_address,
                      code_token_to_send, max_amount_to_send, issuer_token_to_send,
                      code_token_to_be_received, amount_to_be_received, issuer_token_to_be_received,
                      payment_path, transaction_memo=''):
    if not is_seed_matching_address(source_account_seed, source_account_address):
        print("The path payment could not be finalized. Either the given source account address or source"
              "account seed are invalid or they do not match.")
        return

    if not is_address_valid(destination_address):
        print('The given destination address is invalid')
        return

    operations = [create_path_payment_op(destination_address,
                                         code_token_to_send, issuer_token_to_send, max_amount_to_send,
                                         code_token_to_be_received, issuer_token_to_be_received, amount_to_be_received,
                                         payment_path, source=source_account_address)]

    response = submit_operations(source_account_seed, operations, transaction_memo)
    # TODO: Deal with the response


def establish_trustline(source_account_address, source_account_seed, destination_account_address,
                        token_code, token_limit, transaction_memo=''):
    """
    This method is used to create a trustline for a given token code and up to the specified limit,
    with the specified destination account address.
    :param str source_account_address: Address of the account from which the funds will be sent.
    :param str source_account_seed: Seed of the account from which the funds will be sent.
    :param str destination_account_address: Destination address.
    :param str token_code: Code of the token to be sent.
    :param str token_limit: Token limit value of the trustline.
    :param str transaction_memo: Text memo to be included in Stellar transaction. Maximum size of 28 bytes.
    """
    if not is_seed_matching_address(source_account_seed, source_account_address):
        print("The trustline could not be established. Either the given source account address or source"
              "account seed are invalid or they do not match.")
        return

    if not is_address_valid(destination_account_address):
        print('The given destination address is invalid')
        return
    if destination_account_address == source_account_address:
        print('Sending change of trust transaction to own address. This is not allowed')
        return
    if not is_transaction_text_memo_valid(transaction_memo):
        print('The maximum size of the text memo is {} bytes'.format(STELLAR_MEMO_TEXT_MAX_BYTES))
        return

    operations = [create_trust_op(destination_account_address,
                                  token_code,
                                  token_limit,
                                  source_account_address)]
    response = submit_operations(source_account_seed, operations, transaction_memo)
    #process_server_payment_response(response)


def submit_operations(account_seed, operations, transaction_memo):
    """
    This method signs the given operations and submits them to the Stellar network
    :param str account_seed: Seed of the account submitting the operations. It is required to sign the transactions.
    :param Operation operations: Operations to be submitted.
    :param str transaction_memo: Text memo to be included in Stellar transaction. Maximum size of 28 bytes.
    :return: Returns a string containing the server response or None if the transaction could not be submitted.
    :rtype: str or None
    """
    try:
        builder = Builder(secret=account_seed)
    except Exception:
        # Too broad exception because no specific exception is being thrown by the stellar_base package.
        # TODO: This should be fixed in future versions
        print("An error occurred (Please check your Internet connection)")
        return None

    builder.add_text_memo(transaction_memo)
    for operation in operations:
        builder.append_op(operation)
    builder.sign()
    try:
        return builder.submit()
    except Exception:
        # Too broad exception because no specific exception is being thrown by the stellar_base package.
        # TODO: This should be fixed in future versions
        print("An error occurred (Please check your Internet connection)")
        return None


def process_server_payment_response(response):
    if response is None:
        return

    if response.get('status') is not None:
        print("Server response status code: {}".format(response.get('status')))
    if response.get('title') is not None:
        print("Server response title: {}".format(response.get('title')))
    if response.get('detail') is not None:
        print("Server response detail: {}".format(response.get('detail')))
    if 'extras' in response and 'result_xdr' in response['extras']:
        unpacked_tx_result = decode_xdr_transaction_result(response['extras']['result_xdr'])
        print_xdr_transaction_result(unpacked_tx_result)
    if 'result_xdr' in response:
        unpacked_tx_result = decode_xdr_transaction_result(response.get('result_xdr'))
        print_xdr_transaction_result(unpacked_tx_result)


def decode_xdr_transaction_result(xdr_string):
    try:
        xdr_bytes = base64.b64decode(xdr_string)
        return Xdr.StellarXDRUnpacker(xdr_bytes).unpack_TransactionResult()
    except TypeError or ValueError:
        print("Error during base64 decoding")
        return None
    except XDRError:
        print("Error during XDR unpacking procedure")
        return None


def print_xdr_transaction_result(unpacked_tx_result):
        payment_result = unpacked_tx_result.result.results[0].paymentResult
        print("Server response operation result: {}".format(
            'Succeeded' if unpacked_tx_result.result.code == StellarXDR_const.txSUCCESS else 'Failed'))
        print('Server response payment result: {} (Code: {})'.format(str(payment_result), payment_result.code))



# 3rd party imports
from stellar_base.operation import *


def create_account_creation_op(destination, starting_balance, source=None):
        """
        This method creates a Stellar account creation operation. This operation can be used
        to create a new Stellar account by transferring a certain amount of funds to it.
        :param str destination: Address of the new account to be created.
        :param str starting_balance: Amount of funds to transfer to the new account.
        :param str source: TODO
        :return: Returns the newly created operation.
        :rtype: CreateAccount
        """
        return CreateAccount(opts={
            'source': source,
            'destination': destination,
            'starting_balance': starting_balance
        })


def create_payment_op(destination, amount, asset_code='XLM', asset_issuer=None, source=None):
    """
    This method creates a Stellar payment operation. This operation can be used
    to transfer funds between two accounts.
    :param str destination: Destination address to which the funds must be sent.
    :param str amount: Amount of funds to transfer.
    :param str asset_code: Code of the asset to be transferred. If none is specified the native asset is chosen.
    :param str asset_issuer: Issuer of the asset.
    :param str source: TODO
    :return: Returns the newly created operation.
    :rtype: Payment
    """
    return Payment(opts={
        'source': source,
        'destination': destination,
        'asset': Asset(asset_code, asset_issuer),
        'amount': amount
    })


def create_path_payment_op(destination, send_code, send_issuer, send_max, dest_code,
                           dest_issuer, dest_amount, path, source=None):
        """
        This method creates a Stellar path payment operation. This operation can be used
        to use the Stellar decentralized exchange (SDEX) to automatically convert the sent
        asset to the destination asset using the specified path. This enable, for example,
        the source account to transfer XLM while the destination account receives €.
        :param str destination:
        :param str send_code:
        :param str send_issuer:
        :param str send_max:
        :param str dest_code:
        :param str dest_issuer:
        :param str dest_amount:
        :param path: TODO
        :param source: TODO
        :return: Returns the newly created operation.
        :rtype: PathPayment
        """
        assets = []
        for p in path:
            assets.append(Asset(p[0], p[1]))

        opts = {
            'source': source,
            'destination': destination,
            'send_asset': Asset(send_code, send_issuer),
            'send_max': send_max,
            'dest_asset': Asset(dest_code, dest_issuer),
            'dest_amount': dest_amount,
            'path': assets
        }
        return PathPayment(opts)


def create_trust_op(destination, code, limit=None, source=None):
    """
    This method creates a Stellar trust operation. This operation can be used
    to create, delete or update an existing trustline.
    :param str destination: Destination address to which the trustline refers.
    :param str code: Code of the asset to which the trustline refers.
    :param str limit: Amount limit of the asset for the trustline.
    :param str source: TODO
    :return: Returns the newly created operation.
    :rtype: ChangeTrust
    """
    return ChangeTrust(opts={
        'source': source,
        'asset': Asset(code, destination),
        'limit': limit
    })




# Local imports
from .geek_stellar_cmd import GeekStellarCmd
from .cli_session import *


def main():
    """
    Entry point of pygeek-stellar tool. Prints the banner, initializes a
    CLI session and starts the pygeek-stellar CMD interpreter.
    """
    print_banner()

    session = init_cli_session()
    if not session:
        print("No session could be initialized. Exiting..")
        return

    cmd = GeekStellarCmd(session)
    print_current_session_details(session)
    cmd.do_help(None)
    cmd.cmdloop()


def print_banner():
    """
    Prints the a pygeek-stellar banner in the Command Line Interface.
    """
    ch = '#'
    length = len(CLI_BANNER_TEXT) + 8
    spaced_text = ' %s ' % CLI_BANNER_TEXT
    banner = spaced_text.center(length, ch)
    print('\n' + ch * length)
    print(banner)
    print(ch * length + '\n')


def print_current_session_details(session):
    """
    Prints information regarding the selected account for a given CLI session.
    :param CliSession session: Current CLI session.
    """
    print('\nThe following account was chosen: {}\n'.format(session.to_str()))


if __name__ == "__main__":
    main()



# System imports
from cmd import Cmd
import os
import shlex
# Local imports
from .stellar_requests import *
from .stellar_queries import *
from .utils.generic import *
from .cli_session import *


class GeekStellarCmd(Cmd):

    def __init__(self, session):
        super(GeekStellarCmd, self).__init__()
        self.session = session
        self.prompt = '> '

    def cmdloop(self, intro=None):
        """
        Extend the base class cmdloop method to call do_quit() when a KeyboardInterrupt
        (ctr+c) is received
        """
        while True:
            try:
                super(GeekStellarCmd, self).cmdloop()
            except KeyboardInterrupt:
                self.do_quit(None)

    def do_current_account(self, args):
        """
        Prints information regarding the current Stellar account being used
        """
        print('{}'.format(self.session.to_str()))

    def do_get_account_balances(self, args):
        """
        Requests the current XLM (Stellar Lumens) balance from the Stellar Horizon server.
        If no account address is specified the account address of the current CLI session
        will be used.
        Usage: get_account_balances {account_address: optional}
        """
        args = shlex.split(args)
        account_address = args[0] if len(args) >= 1 else self.session.account_address

        balances = get_account_balances(account_address)
        if balances is None:
            print('No balances could be retrieved')
            return
        for balance in balances:
            print('  {}: {}'.format(balance[0], balance[1]))

    def do_get_account_payments(self, args):
        """
        Queries the Stellar Horizon server regarding payments related info.
        If no account address is specified the account address of the current CLI session
        will be used.
        Usage: get_account_payments {account_address: optional}
        """
        args = shlex.split(args)
        account_address = args[0] if len(args) >= 1 else self.session.account_address

        payments = get_account_payments(account_address)
        if payments is None:
            print('No payments could be retrieved')
            return
        print(payments)
        # TODO: Process payments json

    def do_get_account_transactions(self, args):
        """
        Queries the Stellar Horizon server regarding transactions related info.
        If no account address is specified the account address of the current CLI session
        will be used.
        Usage: get_account_transactions {account_address: optional}
        """
        args = shlex.split(args)
        account_address = args[0] if len(args) >= 1 else self.session.account_address

        transactions = get_account_transactions(account_address)
        if transactions is None:
            print('No transactions could be retrieved')
            return
        print(transactions)
        # TODO: Process transactions json

    def do_request_funds(self, args):
        """
        Requests funds from the Stellar Testnet Friendbot. This request will only be successful
        a single time for each Stellar Testnet address. If no account address is specified the
        account address of the current CLI session will be used.
        Usage: request_funds( {account_address: optional}
        """
        args = shlex.split(args)
        account_address = args[0] if len(args) >= 1 else self.session.account_address

        result = fund_using_friendbot(account_address)
        print('Friendbot funding result: ' + result)

    def do_send_donation(self, args):
        """
        Sends a XLM donation as a form of contribution to the development of pygeek-stellar
        Usage: send_donation {amount} {transaction_memo: optional}
        """
        args = shlex.split(args)
        if len(args) < 1:
            print('A XLM amount is mandatory')
            return

        amount = args[0].replace(',', '.')
        memo = '' if len(args) < 2 else args[1]  # memo is optional

        if not is_float_str(amount):
            print('The transfer amount to transfer must but a valid value')
            return

        seed = self.session.fetch_valid_seed()
        if seed is None:
            print("The donation cannot be done without a valid account seed")
            return

        send_payment(self.session.account_address, seed, STELLAR_DONATION_ADDRESS, 'XLM', amount, None, memo)

    def do_create_new_account(self, args):
        """
        Creates a new Stellar address by transferring a given amount of XLM to that address.
        Usage: create_new_account {account_address} {amount} {transaction_memo: optional}
        """
        args = shlex.split(args)
        if len(args) < 2:
            print('An account address and a transfer amount are mandatory')
            return

        address = args[0]
        amount = args[1].replace(',', '.')
        memo = '' if len(args) < 3 else args[2]  # memo is optional

        if not is_float_str(amount):
            print('The transfer amount to transfer must but a valid value')
            return

        seed = self.session.fetch_valid_seed()
        if seed is None:
            print("The account cannot be created without a valid account seed")
            return
        
        create_new_account(self.session.account_address, seed, address, amount, memo)
        
    def do_send_token_payment(self, args):
        """
        Sends a XLM payment to the given destination address
        Usage: send_xlm_payment {destination_address} {token_name} {amount} {transaction_memo: optional}
        """
        args = shlex.split(args)
        if len(args) < 3:
            print('A destination address, token name and a transfer amount are mandatory')
            return

        destination = args[0]
        token_name = args[1]
        amount = args[2].replace(',', '.')
        memo = '' if len(args) < 4 else args[3]  # memo is optional

        if not is_float_str(amount):
            print('The transfer amount to transfer must but a valid value')
            return

        seed = self.session.fetch_valid_seed()
        if seed is None:
            print("The token payment cannot be done without a valid account seed")
            return

        # TODO: Check who should be the issuer! Should it be a input from the user?
        send_payment(self.session.account_address, seed, destination, token_name, amount,
                     self.session.account_address, memo)

    def do_send_xlm_payment(self, args):
        """
        Sends a XLM payment to the given destination address
        Usage: send_xlm_payment {destination_address} {amount} {transaction_memo: optional}
        """
        args = shlex.split(args)
        if len(args) < 2:
            print('A destination address and a transfer amount are mandatory')
            return

        destination = args[0]
        amount = args[1].replace(',', '.')
        memo = '' if len(args) < 3 else args[2]  # memo is optional

        if not is_float_str(amount):
            print('The transfer amount to transfer must but a valid value')
            return

        seed = self.session.fetch_valid_seed()
        if seed is None:
            print("The payment cannot be done without a valid account seed")
            return

        send_payment(self.session.account_address, seed, destination, 'XLM', amount, None, memo)

    def do_establish_trustline(self, args):
        """
        Creates a new Stellar token with the given name and limit
        Usage: establish_trustline {destination_address} {token_name} {token_limit}
        """
        args = shlex.split(args)
        if len(args) < 3:
            print('A destination address, token name and token limit are mandatory')
            return

        destination = args[0]
        token_name = args[1]
        token_limit = args[2]

        if not is_int_str(token_limit):
            print('The token limit must but a valid integer value')
            return

        seed = self.session.fetch_valid_seed()
        if seed is None:
            print("The trustline cannot be established without a valid account seed")
            return

        establish_trustline(self.session.account_address, seed, destination, token_name, token_limit)

    @staticmethod
    def do_cls(args):
        """
        Clear the screen
        """
        os.system('cls' if os.name == 'nt' else 'clear')  # Use cls for Windows and clear for Unix

    @staticmethod
    def do_quit(args):
        """
        Quits the program.
        """
        print("Quitting.")
        raise SystemExit




from setuptools import setup

setup(name='pygeek-stellar',
      version='0.2',
      description='A Python CLI to interact with the Stellar network.',
      url='https://github.com/XavierAraujo/pygeek-stellar',
      author='Xavier Araújo',
      author_email='xavier.araujo92@gmail.com',
      license='MIT',
      packages=['pygeek_stellar', 'pygeek_stellar.utils'],
      install_requires=[
            'stellar-base',
            'cryptography',
            'requests'
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      scripts=['bin/pygeek-stellar'],
      python_requires='>=3.7',
      zip_safe=False)



# 3rd party imports
import stellar_base.utils
from stellar_base.exceptions import *
from stellar_base.keypair import Keypair
from stellar_base.address import Address

STELLAR_MEMO_TEXT_MAX_BYTES = 28


def is_address_valid(address):
    """
    Checks if a given Stellar address is valid. It does not check if it exists on the Stellar
    network, only if it is correctly formatted.
    :param str address: address to be evaluated.
    :return: Returns true if the given address is valid and false otherwise.
    :rtype: bool
    """
    if address is None:
        return False
    try:
        stellar_base.utils.decode_check('account', address)
        return True
    except DecodeError:
        return False


def is_seed_valid(key):
    """
    Checks if a given Stellar seed is valid.
    :param str key: Seed to be evaluated.
    :return: Returns true if the seed is valid and false otherwise.
    :rtype: bool
    """
    if key is None:
        return False
    try:
        stellar_base.utils.decode_check('seed', key)
        return True
    except DecodeError:
        return False


def is_transaction_text_memo_valid(memo):
    """
    Checks if a given Stellar transaction text memo is valid. To be valid the text memo
    can only have, at most, 28 bytes.
    :param str memo: Text memo to be evaluated.
    :return: Returns true if the given text memo is valid and false otherwise.
    :rtype: bool
    """
    if memo is None:
        return False
    return False if len(memo) > STELLAR_MEMO_TEXT_MAX_BYTES else True


def is_seed_matching_address(seed, address):
    """
    Checks if the specified seed address matches the specified address.
    :param str seed: Seed to be evaluated.
    :param str address: Address to be evaluated.
    :return: Returns true if seed address matches the specified address, and false otherwise.
    :rtype: bool
    """
    if not is_seed_valid(seed) \
            or not is_address_valid(address):
        return False

    keypair = Keypair.from_seed(seed=seed)
    if keypair.address().decode() == address:
        return True
    return False


def is_account_existent(address):
    """
    Checks if a given Stellar address exists in the network. It assumes that the address
    parameter received is a valid address string.
    :param str address: address to be evaluated.
    :return: Returns true if the given address exists in the network and false otherwise.
    :rtype: bool
    """
    return True if get_address_details_from_network(address) is not None else False


def get_address_details_from_network(address):
    """
    Queries the Stellar network regarding the details of the specified account address.
    :param str address: address to be evaluated.
    :return: In case of success returns a Stellar Address object with the updated address information, fetched from
    the Stellar network. In case of failure returns None
    :rtype: Address or None
    """
    if not is_address_valid(address):
        print('Trying to get information of an invalid address.')
        return None

    try:
        address = Address(address=address)
        address.get()  # Get the latest information from Horizon
    except AccountNotExistError:
        print('The specified account does not exist.')
        return None
    except HorizonError:
        print('A connection error occurred (Please check your Internet connection).')
        return None
    return address



# System imports
from pathlib import Path

CLI_BANNER_TEXT = 'PYGEEK-STELLAR'

# Config file related constants
DEFAULT_CONFIG_FILE = '{}/.pygeek-stellar.config'.format(str(Path.home()))
JSON_ACCOUNTS_TAG = 'accounts'
JSON_ACCOUNT_NAME_TAG = 'account_name'
JSON_ACCOUNT_ADDRESS_TAG = 'address'
JSON_ACCOUNT_SEED_TAG = 'secret_seed'
CONFIG_FILE_EMPTY_JSON = {JSON_ACCOUNTS_TAG: []}

# Stellar related constants
STELLAR_HORIZON_TESTNET_URL = 'https://horizon-testnet.stellar.org'
STELLAR_ASSET_TYPE_XLM = 'native'
STELLAR_DONATION_ADDRESS = 'GBLHVU7EJMSUW72PINTRBRIHT55ZQ7HXFAXELG2JG53X4VAZSHZKLEY6'
